# Copyright (C) 2005 Ales Smrcka
# Use of this software is subject to the terms of the GNU General
# Public License
#

from xml.dom import minidom, Node
import re

def parsepath(path):
	"""internal use only: parsepath(path) - transform path like 'node/subnode(attr1=value1,attr2=value2)[idx]{el.value}/subsubnode' into processible structure (element value is not supported at the moment)"""
	# remove first and last '/'
	if path[0] == '/': path=path[1:]
	if path[-1] == '/': path=path[:-1]
	elements = []
	while path:
		where=0 # state in finite automaton
		element=''
		attrs={}
		idx=''
		attrname=''
		attrvalue=''
		value=''
		for i in range(len(path)):
			c = path[i]
			if where==0: # inside element
				if c == '/': where=8
				elif c == '(': where=1
				elif c == '[': where=4
				elif c == '{': where=6
				else: element += c
			elif where==1: # inside attribute name
				if c == '=': where=2
				else: attrname += c
			elif where==2: # inside attribute value
				if c == ',' or c == ')':
					attrs[attrname] = attrvalue
					attrname=''
					attrvalue=''
				if c == ',': where=1
				elif c == ')': where=3
				else: attrvalue += c
			elif where==3: # before element index
				if c == '/': where=8
				elif c == '[': where=4
				elif c == '{': where=6
			elif where==4: # inside element index
				if c == ']': idx=int(idx); where=5
				else: idx += c
			elif where==5: # before element value
				if c == '/': where=8
				elif c == '{': where=6
			elif where==6: # inside element value
				if c == '}': where=7
				elif c == '\\': where=11
				else: value += c
			elif where==11: # inside element value, special char.
				if c == '}':
					value += '}'
				else:
					value += '\\'+c
				where = 6
			elif where==7: # after element value
				if c == '/': where=8
			elif where==8: # next element definition
				i -= 1
				break
		if idx == '': idx=-1
		elements.append((element, attrs, idx, value))
		path = path[i+1:]
	return elements

def element_childs(node, nextelement):
	"""internal use only: node (xml.dom.Node), nextelement: ('name',{attrs values},idx)"""
	if nextelement[0] == '..':
		if node.parentNode:
			return [node.parentNode]
		else:
			return [] 
	results=[]
	i=0
	for c in node.childNodes:
		# child is element type and name matches
		if c.nodeType == Node.ELEMENT_NODE and c.nodeName == nextelement[0]:
			# check for attribute values
			attrok=True
			for (name, value) in list(nextelement[1].items()):
				try:
					# try look for an attribute
					if c.getAttribute(name) != value:
						attrok=False
						break
				except:
					# attribute does not exist
					attrok=False
					break
			# attributes are ok
			if attrok:
				# index of element is fine
				if nextelement[2] == -1 or i == nextelement[2]:
					valueok=False
					if nextelement[3]:
						e = xmlelement(c)
						if re.search(nextelement[3], e.value):
							valueok=True
					else:
						valueok=True
					if valueok:
						# all is ok
						results.append(c)
						
				i += 1
	return results

class SimpleXMLError(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)
		
class xmldoc:
	def __init__(self, filename=""):
		self.filename=""
		self.dom=None
		self.__dict__['root']=None
		if filename:
			self.load(filename)
	
	def __bool__(self):
		return self.dom == None
	
	def __setattr__(self, name, value):
		if name == 'root':
			self.new_root(value)
		else:
			self.__dict__[name] = value

	def load(self, filename):
		"""load(filename) - load source XML data from filename"""
		self.dom = minidom.parse(filename)
		self.__dict__['root'] = xmlelement(self.dom.documentElement)
		self.filename = filename

	def load_string(self, string):
		"""loadString(string) - read source XML data from string"""
		self.dom = minidom.parseString(string)
		self.__dict__['root'] = xmlelement(self.dom.documentElement)
		
	def save(self, filename=''):
		"""save(filename='') - save XML data into filename"""
		if not filename and not self.filename:
			raise SimpleXMLError("You must select filename")
		if not filename and self.filename:
			filename = self.filename
		f=open(filename, 'w')
		f.write(self.toxml())
		f.write('\n')
		f.close()

	def savepretty(self, filename=''):
		"""savepretty(filename='') - save pretty XML data into filename (warning: changes white characters of data and merges data inside one element"""
		if not filename and not self.filename:
			raise SimpleXMLError("You must select filename")
		if not filename and self.filename:
			filename = self.filename
		f=open(filename, 'w')
		f.write(self.toprettyxml())
		f.write('\n')
		f.close()

	def toxml(self):
		"""toxml() - return XML data as string"""
		return self.dom.toxml('utf-8')
	
	def _stripvalues(self, node):
		element = xmlelement(node)
		element.value = element.value
		del element
		for child in node.childNodes:
			if child.nodeType == Node.ELEMENT_NODE:
				self._stripvalues(child)

	def toprettyxml(self):
		"""toprettyxml() - return XML data as pretty string (warning: changes white characters of data and merges data inside one element"""
		self.root.value = self.root.value
		self._stripvalues(self.root.element)
		return self.dom.toprettyxml(encoding='utf-8')
		
	def new_root(self, name):
		"""new_root(name) - set new root element. All previous elements will be deleted"""
		if self.dom:
			if self.dom.documentElement:
				self.dom.removeChild(self.dom.documentElement)
		else:
			self.dom = minidom.Document()
		self.__dict__['root'] = xmlelement(name, self.dom)
		self.dom.appendChild(self.root.element)
		return self.root

	def __getitem__(self, path):
		"""self[path] - return xmlelement object appropriate to path"""
		elements=self.elements(path)
		if len(elements)==0:
			return None
		elif len(elements)==1:
			return elements[0]
		else:
			return elements
		
	def elements(self, path):
		"""element(path) - return xmlelement object appropriate to path. path must be unicode string if it contains any non ascii characters"""
		epath = parsepath(path)
		if epath[0][0] != self.dom.documentElement.nodeName:
			raise SimpleXMLError("Wrong path was set - root element does not match")
		nodes=[self.dom.documentElement]
		del epath[0] # delete root element and start with next elements
		for e in epath:
			childnodes=[]
			for n in nodes:
				for c in element_childs(n, e):
					if c not in childnodes:
						childnodes.append(c)
			prev=None
			
			nodes = childnodes
		return list(xmlelement(n) for n in nodes)
		
class xmlelement:
	def __init__(self, node, document=None):
		"""__init__(node, document=None) - node can be either minidom.Node type or string, which creates new (element type) node with document owner set to document (minidom.Document)"""
		if type(node) == str or type(node) == str:
			if not document:
				raise SimpleXMLError("You must specify the document owner")
			self.element = minidom.Element(node)
			self.element.ownerDocument = document
			self.__dict__['name'] = node
		elif node.nodeType != Node.ELEMENT_NODE:
			raise SimpleXMLError("node must be type of Node.ELEMENT_NODE")
		else:
			self.element = node
			self.__dict__['name'] = self.element.nodeName
		self._updatedata()	# set self.data and self.value
		
	def __bool__(self):
		return True
		
	def __setattr__(self, name, value):
		if name == 'value':
			self._set_value(value)
		elif name == 'name':
			self.element.tagName = value
			self.__dict__['name'] = value
		else:
			self.__dict__[name] = value

	def __getitem__(self, path):
		"""self[path] - return xmlelement object appropriate to path"""
		elements=self.elements(path)
		if len(elements)==0:
			return None
		elif len(elements)==1:
			return elements[0]
		else:
			return elements

	def __str__(self):
		return self.value
	
	def _strip_value(self):
		self.__dict__['value'] = ''
		prev = ''
		for c in self.data.strip():
			if c.isspace() and not (prev.isspace() or not prev):
				self.__dict__['value'] += ' '
			elif not c.isspace():
				self.__dict__['value'] += c
			prev = c
		return self.value

	def _updatedata(self):
		# create and set self.data with all data inside this element
		# also create and set self.value as brute-stripped self.data
		self.__dict__['data'] = ''
		for node in self.element.childNodes:
			if node.nodeType in (Node.TEXT_NODE, Node.CDATA_SECTION_NODE):
				self.__dict__['data'] += node.data
		self._strip_value()
		return self.data

	def _set_value(self, value):
		"""_set_value(value) - sets new value (data) of element. value must be unicode string if it contains any non ascii characters"""
		# at first: delete all text child nodes
		for node in self.element.childNodes:
			if node.nodeType in (Node.TEXT_NODE, Node.CDATA_SECTION_NODE):
				self.element.removeChild(node)
		textNode = minidom.Text()
		textNode._set_data(str(value))
		self.element.insertBefore(textNode, self.element.firstChild)
		self._updatedata()
		return self.value
	
	def attr(self, name, value=None):
		"""attr(name, value=None) - get or set value of attribute. name or value must be unicode strings if it contain any non ascii characters"""
		if value == None:
			return self.element.getAttribute(name)
		else:
			return self.element.setAttribute(name, value)
	
	def removeattr(self, name):
		"""removeattr(name) - remove attribute. name must be unicode string if it contains any non ascii characters"""
		self.element.removeAttribute(name)
		
	def elements(self, path):
		"""element(path) - return list of xmlelements appropriate to path. path must be unicode string if it contains any non ascii characters"""
		epath = parsepath(path)
		nodes=[self.element]
		for e in epath:
			childnodes=[]
			for n in nodes:
				childnodes += element_childs(n, e)
			nodes = childnodes
		return list(xmlelement(n) for n in nodes)

	def toxml(self):
		"""toxml() - return XML form as string"""
		return self.element.toxml('utf-8')

	def newchild(self, elementname):
		"""newchild(name) - create and return new child element. name must be unicode string if it contains any non ascii characters"""
		child = xmlelement(elementname, self.element.ownerDocument)
		self.element.appendChild(child.element)
		return child

	def appendchild(self, element):
		"""appendchild(element) - append new child xmlelement"""
		self.element.appendChild(element.element)

	def insertbefore(self, newelement, refelement):
		"""insertbefore(newelement, refelement) - insert child element before refelement\nand remove it from its old position if any"""
		newelement.remove()
		self.element.insertBefore(newelement.element, refelement.element)
	
	def movebefore(self, refelement):
		"""movebefore(refelement) - move this element before refelement"""
		self.remove()
		self.element.insertBefore(self.element, refelement.element)
		
	def moveafter(self, refelement):
		"""movebefore(refelement) - move this element before refelement"""
		self.remove()
		self.element.insertBefore(self.element, refelement.element)
	
	def remove(self):
		"""remove() - remove this element from its parent"""
		if self.element and self.element.parentNode:
			self.element.parentNode.removeChild(self.element)

